import a2.plotting.analysis
import a2.plotting.figures
import a2.plotting.histograms
import a2.plotting.parallel_plotting
import a2.plotting.timeseries
import a2.plotting.utils_plotting
import a2.plotting.weather_maps
